#!/usr/bin/env python
# -*- coding: utf-8 -*-

# forbi: a TCP-based communication tool
# Copyright (C) 2010  Niels Serup

# This file is part of forbi.
#
# forbi is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# forbi is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General
# Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with forbi. If not, see
# <http://www.gnu.org/licenses/>.

##[ Maintainer  ]## Niels Serup <ns@metanohi.org>
##[ Description ]## Functions for maintaining a list of connections

import forbi.misc as misc

class ConnectionInfo(misc.Container):
    """Info about a connection"""
    def __init__(self, **kwargs):
        # All of these arguments are not meant to be filled at once.
        self.addr = kwargs.get('addr')
        self.user = kwargs.get('user')
        self.port = kwargs.get('port')
        self.id = kwargs.get('id')
        self.sender = kwargs.get('sender')

class ConnectionsList(misc.Container):
    """A list of connections with info"""
    def __init__(self):
        self.connections = []

    def add(self, **kwargs):
        """Add a connection with as many or as few arguments as you want"""
        new_info = ConnectionInfo(**kwargs)
        self.connections.append(new_info)
        return new_info

    def remove(self, conn):
        """Remove a connection"""
        assert conn in self.connections, \
            'connection object not in connections list'
        self.connections.remove(conn)

    def get_users(self):
        """Get all usernames"""
        return set(x.user for x in self.connections if x.user is not None)

    def find(self, **kwds):
        """
        Find the first connection whose items match all of the
        keyword arguments
        """
        for conn in self.connections:
            ok = True
            for key, val in kwds.iteritems():
                if conn.__dict__[key] != val:
                    ok = False
                    break
            if ok:
                return conn



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# forbi: a TCP-based communication tool
# Copyright (C) 2010  Niels Serup

# This file is part of forbi.
#
# forbi is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# forbi is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General
# Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with forbi. If not, see
# <http://www.gnu.org/licenses/>.

##[ Maintainer  ]## Niels Serup <ns@metanohi.org>
##[ Description ]## What to do when a message is received

import Tkinter as tk
import tkFont
import forbi.misc as misc

# Global variables
program = None
windows = None
new_message = None


# Internal way of showing new messages
class MessageWindows(object):
    """A hidden root window with dynamically addable visible subwindows"""
    def __init__(self):
        self.queue = []
        self.created = False

    def start(self):
        """In a new thread, create the root window and run the event loop"""
        misc.QuickThread(self.create)
        while not self.created:
            pass

    def create(self):
        """Create an empty root window"""
        self.root = tk.Tk()
        self.root.withdraw()
        self.root.bind('<<show_message>>', self.show_message)
        self.created = True
        tk.mainloop()

    def new_message(self, username, msg):
        """Add a new window"""
        self.queue.append((username, msg))
        self.root.event_generate('<<show_message>>', when='tail')

    def show_message(self, *args):
        """Format and show the new message"""
        username, msg = self.queue.pop(0)

        top = tk.Toplevel()
        top.geometry('320x240')
        top.maxsize(320, 240)
        top.title('Message received')

        top.bind(sequence='<Control-q>', func=lambda *args: top.destroy())

        scrollbar = tk.Scrollbar(top)
        text = tk.Text(top)
        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        text.pack(side=tk.LEFT, fill=tk.Y)
        scrollbar.config(command=text.yview)
        text.config(yscrollcommand=scrollbar.set)

        user_font = tkFont.Font(size=14, weight='bold')
        text.tag_config('u', background='black', foreground='white', font=user_font)

        text.insert(tk.END, username + '\n', ('u',))
        text.insert(tk.END, msg)

def _new_msg_gui(user_from, user_to, msg):
    global windows
    windows.new_message(user_from, msg)


# External way of showing new messages
def _new_msg_program(user_from, user_to, msg):
    global program
    misc.exec_program(program, user_from, user_to, msg)

def init(prog=None):
    """Prepare this module for new messages"""
    global new_message
    if prog is None:
        # Prepare the GUI
        global windows, _new_msg_gui
        windows = MessageWindows()
        windows.start()
        new_message = _new_msg_gui
    else:
        # Prepare to call an external program
        global program, _new_msg_program
        program = prog
        new_message = _new_msg_program

if __name__ == '__main__':
    # For testing it
    win = MessageWindows()
    win.start()
    win.new_message(u'Test user', u'\n'.join((u'Hello!' for i in xrange(25))))
    win.new_message(u'Test user2', u'\n'.join((u'!olleH' for i in xrange(52))))
    while True: pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# forbi: a TCP-based communication tool
# Copyright (C) 2010  Niels Serup

# This file is part of forbi.
#
# forbi is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# forbi is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General
# Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with forbi. If not, see
# <http://www.gnu.org/licenses/>.

##[ Maintainer  ]## Niels Serup <ns@metanohi.org>
##[ Description ]## Functions for keeping data

try:
    import cPickle as pickle
except ImportError:
    import pickle
import hashlib
import forbi.misc as misc

class UserSaves(misc.Container):
    """
    Contains usernames together with hashed passwords. Support for
    loading and saving.
    """
    def __init__(self, filename):
        self.path = filename
        self.users = {}

    def load(self):
        """Load data from a file"""
        try:
            with open(self.path, 'rb') as f:
                users = pickle.load(f)
                for user, password in users.iteritems():
                    if user not in self.users:
                        self.users[user] = password
        except IOError:
            pass

    def clear(self):
        self.users = {}

    def save(self):
        """Save data to a file"""
        try:
            with open(self.path, 'wb') as f:
                pickle.dump(self.users, f)
        except IOError:
            pass

    def add(self, user, password):
        """Add a user"""
        assert not self.has(user), 'User %s already exists' % user
        self.set(user, password)

    def change(self, user, password):
        """Change a user's password"""
        assert self.has(user), 'User %s does not exist' % user
        self.set(user, password)

    def set(self, user, password):
        """Set a user's password"""
        self.users[user] = self.hash(password)
        self.save()

    def remove(self, user):
        """Remove a user"""
        assert user in self.users, 'User %s does not exist' % user
        del self.users[user]
        self.save()

    def hash(self, text):
        """Hash digest text"""
        return hashlib.sha256(text).digest()

    def has(self, user):
        """See if user is saved"""
        return user in self.users.iterkeys()

    def match(self, user, password):
        """See if password matches user's password"""
        assert user in self.users, 'User %s does not exist' % user
        return self.users[user] == self.hash(password)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# forbi: a TCP-based communication tool
# Copyright (C) 2010  Niels Serup

# This file is part of forbi.
#
# forbi is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# forbi is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General
# Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with forbi. If not, see
# <http://www.gnu.org/licenses/>.

##[ Maintainer  ]## Niels Serup <ns@metanohi.org>
##[ Description ]## Server communicator

from forbi.communicator import ForbiCommunicator, \
    ForbiRequestHandler, ForbiSocket
import forbi.misc as misc
import forbi.usersaves as usersaves
import forbi.connlist as connlist
import forbi.crypt as crypt
import socket
import time

class ForbiServer(ForbiCommunicator):
    """
    A server capable of receiving and sending messages from and to an
    arbitrary number of sources independently
    """
    def __init__(self, port=54547, host='', **kwds):
        self.port = port
        self.host = host

        self.default_client_port = kwds.get(u'default_client_port') or \
            self.port

        self.user_saves = usersaves.UserSaves(kwds.get(u'save_file'))
        self.connections = connlist.ConnectionsList()

        # Create the server handler
        ForbiCommunicator.__init__(
            self, self.port, '', ForbiServerRequestHandler,
            kwds.get(u'runner'))

    def start(self):
        """Start the server"""
        self.set_output_prefix(u'server')
        ForbiCommunicator.start(self)
        self.port = self.serving_port
        self.user_saves.load()
        self.run()

class ForbiServerRequestHandler(ForbiRequestHandler):
    """
    A request handler that creates a new socket back to the source
    whenever a new connection is made
    """
    def start(self):
        """Finalize the connection"""
        self.conn_info = \
            self.parent.connections.add(addr=self.client_address)
        self.addr_text_recv = repr(self.conn_info.addr)
        self.ping_thread = None

    def end(self):
        # Terminate ping thread
        if self.ping_thread is not None:
            self.ping_thread.terminate()
        # Remove this connection from list of connections
        if self.conn_info.user is not None:
            self.state(u'%s logged out' % repr(self.conn_info.user)[1:])
        self.parent.connections.remove(self.conn_info)
        self.end = misc.nothing # If it's called again

    def connect_back(self):
        """Create a connection back to the source connector"""
        self.state(u'creating socket to %s...' % \
                       repr(self.client_address[0]))
        self.addr_text_send = repr((self.conn_info.addr[0],
                                    self.conn_info.port))
        self.sender = ForbiSocket(self.conn_info.addr[0],
                                  self.conn_info.port,
                                  self.parent.priv_key)
        self.conn_info.sender = self.sender
        if self.do_crypt:
            self.sender.pub_key = self.pub_key
            self.sender.do_crypt = True

        self.state(u'address %s uses %d as server port' %
                   (self.addr_text_recv, self.conn_info.port))
        self.state(u'sending identifier string back to %s' % self.addr_text_send)
        ans = self.send('id ' + self.conn_info.id, False)
        if ans == u'ok':
            self.state(u'%s has accepted identifier string' %
                       self.addr_text_send)
        else:
            self.error(u'%s has not accepted identifier string' %
                       self.addr_text_send)
            raise misc.ForbiError
        self.ping_thread = misc.BasicQuickThread(self.ping)
        self.ping_thread.start()

    def ping(self):
        """Ping source connector with an interval of 256 seconds"""
        self.state(u'started pinging %s.' % self.addr_text_send)
        while True:
            if self.ping_thread.do_terminate:
                break
            try:
                self.state(u'pinging %s.' % self.addr_text_send)
                if self.send(u'ping') != u'pong':
                    self.error(u'did not receive a pong from %s.'
                               % self.addr_text_send)
                    break
                time.sleep(256)
            except Exception:
                break
        self.state(u'stopped pinging %s.' % self.addr_text_send)
        try:
            self.request.shutdown(socket.SHUT_RDWR)
        except socket.error:
            pass

    def send(self, msg, utf8=True):
        """Send a message"""
        return self.sender.send(msg, utf8)

    def receive(self, msg):
        """Receive a message and act on it"""
        if msg.startswith('public_key '):
            # Receive public key from client and send own public key back
            pub_key_str = msg.split(' ', 1)[1]
            pub_key = crypt.create_pub_key_from_string(pub_key_str)
            self.pub_key = pub_key
            self.answer(self.parent.pub_key_str, False)
            self.do_crypt = True
            self.state(u'now using encryption')

        elif msg.startswith(u'version '):
            # Receive client version and compare with own version
            client_version = msg.split(u' ', 1)[1]
            self.answer(misc.version_num_text)
            if tuple(map(int, client_version.split(u'.'))) > \
                    misc.version_num:
                self.error(u'incompatible versions with %s' % self.addr_text_recv)
                return True
            self.state(u'client at %s using version %s' %
                       (self.addr_text_recv, client_version))
            self.state(u'server using version %s' % misc.version_num_text)
            self.state(u'versions compatible')

        elif msg.startswith('id '):
            # Receive random id string and save for later when
            # identification is needed
            self.state(u'received identifier string from %s' % self.addr_text_recv)
            self.conn_info.id = msg.split(' ', 1)[1]
            self.answer(u'ok')

        elif msg.startswith(u'port'):
            # Receive client listening port number and create a
            # connection to it
            self.state(u'received port number from %s' % self.addr_text_recv)
            if u' ' in msg:
                port = int(msg.split(u' ', 1)[1])
            else:
                port = self.parent.default_client_port
            self.conn_info.port = port
            self.answer(u'ok')
            self.connect_back()

        elif msg.startswith(u'register '):
            # Attempt to register the user
            user, password = msg.split(u' ', 1)[1].split(u'\t')
            try:
                self.parent.user_saves.add(user, password)
                self.answer(u'ok')
            except AssertionError:
                self.answer(u'error 1')

        elif msg.startswith(u'login '):
            # Attempt to login
            user, password = msg.split(u' ', 1)[1].split(u'\t')
            if user in self.parent.connections.get_users():
                self.answer(u'error 1')
            elif not self.parent.user_saves.has(user):
                self.answer(u'error 3')
            elif not self.parent.user_saves.match(user, password):
                self.answer(u'error 2')
            else:
                self.state(u'%s logged in' % repr(user)[1:])
                self.conn_info.user = user
                self.answer(u'ok')

        elif msg == 'logout':
            # Attempt to logout
            if self.conn_info.user is not None:
                self.state(u'%s logged out' % repr(self.conn_info.user)[1:])
                # The only way to tell if there is someone logged in
                # from this connection is to look at self.conn_info.user
                self.conn_info.user = None
                self.answer(u'ok')
            else:
                self.answer(u'error no one logged in')

        elif msg == 'unregister':
            # Attempt to unregister (and logout at the same time)
            if self.conn_info.user is not None:
                self.parent.user_saves.remove(self.conn_info.user)
                self.state(u'%s unregistered' % repr(self.conn_info.user)[1:])
                self.conn_info.user = None
                self.answer(u'ok')
            else:
                self.answer(u'error no one logged in')

        elif msg == u'exit':
            # Client shuts down, end this thread
            self.end()
            self.answer(u'ok')
            return True

        elif msg == u'users':
            # Send a list of users back (but only if the client is
            # logged in)
            if self.conn_info.user is not None:
                self.answer(u'ok ' + u'\t'.join(sorted(self.parent.connections.get_users())))
            else:
                self.answer(u'error')

        elif msg.startswith(u'message '):
            # Attempt to forward the message
            user_to, mesg = msg.split(u' ', 1)[1].split(u'\t')
            user_from = self.conn_info.user
            self.state(u'message from %s to be sent to %s' %
                       (repr(user_from)[1:], repr(user_to)[1:]))
            conn = self.parent.connections.find(user=user_to)
            if conn is None:
                self.answer(u'error user does not exist')
            else:
                self.answer(u'ok')
                conn.sender.send(u'message %s\t%s' % (user_from, mesg))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# forbi: a TCP-based communication tool
# Copyright (C) 2010  Niels Serup

# This file is part of forbi.
#
# forbi is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# forbi is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General
# Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with forbi. If not, see
# <http://www.gnu.org/licenses/>.

##[ Maintainer  ]## Niels Serup <ns@metanohi.org>
##[ Description ]## Settings parser

import os
try:
    from qvikconfig import parse as config_parse
except ImportError:
    from forbi.external.qvikconfig import parse as config_parse

basic_config_translations = {}

class SettingsParser(object):
    """
    A parser of options and settings with support for loading data
    from config files
    """
    def __init__(self, ok_config_translations={}, **etc):
        for key, val in basic_config_translations.iteritems():
            ok_config_translations[key] = val
        self.ok_config_translations = ok_config_translations
        self.ok_config_values = set(ok_config_translations.keys())

        for x in etc.items():
            self.__setattr__(*x)

    def is_nil(self, prop):
        return not prop in dir(self) or self.__dict__[prop] is None

    def set_if_nil(self, prop, val):
        if self.is_nil(prop):
            self.__dict__[prop] = val

    def load_from_file(self):
        if not self.is_nil('config_file_path'):
            # Attempt to parse a config file
            try:
                conf = config_parse(self.config_file_path)
                ok_keys = set(conf) & self.ok_config_values

                for key in ok_keys:
                    o_key = key
                    if key in self.ok_config_values:
                        n_key = self.ok_config_translations[key]
                        if n_key is not None:
                            key = n_key
                        self.set_if_nil(key, conf[o_key])
            except IOError:
                pass



#!/usr/bin/env python

# qvikconfig: read and write simple config files
# Copyright (C) 2010  Niels Serup

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Version:...... 0.1.0
# Maintainer:... Niels Serup <ns@metanohi.org>
# Website:...... http://metanohi.org/projects/qvikconfig/
# Development:.. http://gitorious.org/qvikconfig

# This is a Python module. It reads and writes simple documentation
# files. The basic syntax is 'property = value'. Documentation is
# included in this very file.

# Imports
import re
import decimal # For floats which would otherwise lose information

# Documentation
__doc__ = """
==========
qvikconfig
==========

qvikconfig is a parser of a simple config file syntax called the
qvikconfig format. The basic syntax is 'property = value'.

qvikconfig offers two main functions: ``parse`` and ``dump``. Use
``parse`` to read config files and ``dump`` to write them.

Documentation
=============

To see an example of qvikconfig config files and how to parse and dump
them, see the files in the ``tests/`` directory of the
distribution. Alternatively, these examples can be downloaded from
<http://metanohi.org/projects/qvikconfig/>.

qvikconfig can be considered a dictionary that can link both single
values and lists of values to named properties. Entries are
line-separated. A simple example::

  # This is a comment.
  name = Example # This is also a comment.
  descriptions = This is an example, Dette er et eksempel

When parsed, this will be translated to this Python dict:

  {'name': 'Example',
  'descriptions': ['This is an example', 'Dette er et eksempel']}

It is, in most cases, optional to enclose strings in
quotes. Whitespace is completely optional.


Lists
-----

qvikconfig uses the comma character to split values, as seen in the
example above. When there is only one value associated with a
property, there will be no list. Otherwise there will be one. To
create a list of only one value, append two commas to your value, like
this::

  a-list=24,,

This returns this dict::

  {'a-list': [24]}


Accepted data
-------------

qvikconfig accepts and understands strings, numbers, True, False, and
None. Everything else normally results in unknown behavior.

Strings are represented using text, sometimes enclosed in
quotes. Numbers are represented using numbers. True, False and None
are represented by true, false and none.


Using more than one line
------------------------

You can spread lists across several lines like this::

  long list = A,
  B, C, D,
  E, F, G,
  H, I, J,
  K, L, M,
  N, P, O

This will, naturally, result in this dict::

  {'long list': ['A', 'B', 'C', 'D', 'E', 'F',
  'G', 'H','I', 'J', 'K', 'L', 'M', 'N', 'P', 'O']}

You can make single values take up more than one line like this::

  a value = The wall had moved away from the boundaries \
of the house.

..which creates this when parsed::

  {'a value': 'The wall had moved away from the boundaries of the house.'}


Escaping text
-------------

Certain strings cannot be represented in qvikconfig config files
without being enclosed in quotes. The strings 'true', 'false' and
'none' must have quotes (both "..." and '...' are accepted), for
example. The same goes for, in the case of values, strings containing
commas and quotes or ending with a backslash, and in the case of
properties, equal signs. While spaces and tabs are just fine, newlines
(and carriage returns and so on) must be enclosed in quotes. Example::

  # The equality sign is not a comma.
  '=' = 'not a ,'

..resulting in::

  {'=': 'not a ,'}


Version information
-------------------

This is version 0.1.0 of qvikconfig. New releases are made available
online at the website mentioned earlier.

Copyright (C) 2010  Niels Serup
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>
This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law.
"""


# Constants
_config_comment_regex = re.compile(r'#.*')

# The parser class
class _QvikParser(dict):
    def __init__(self, in_file=None, **kwargs):
        # Try to find out if input is a file or a string. If it's a
        # file, extract the data.
        
        def fil2dat(t):
            # If it's a file..
            try:
                return open(t, 'U').read()
            except TypeError:
                try:
                    return t.read().replace(
                        '\r\n', '\n').replace('\r', '\n')
                except Exception:
                    raise TypeError('argument not file')

        # Testing
        fil = kwargs.get('file')
        if not fil:
            dat = kwargs.get('data')
            if not dat:
                if in_file:
                    fil = in_file
                    dat = fil2dat(fil)
                else:
                    raise TypeError('missing arguments')
            else:
                if not isinstance(dat, basestring):
                    raise TypeError('data must be a string')
        else:
            dat = fil2dat(fil)

        # Save the data in case someone needs it
        self.rawdata = dat
        # Parse
        self.parse_data()

    def convert(self, dat):
        # Convert to internal Python data
        if dat.startswith('"') or dat.startswith("'"):
            return eval(dat)
        if dat == 'true':
            return True
        if dat == 'false':
            return False
        if dat == 'none':
            return None
        try:
            return int(dat)
        except ValueError:
            try:
                return decimal.Decimal(dat)
            except decimal.InvalidOperation:
                return dat

    def parse_data(self):
        # Parse the rawdata string
        data = _config_comment_regex.sub('', self.rawdata).split('\n')
        data.append('')

        for i in range(len(data) - 1):
            # All entries have one property and one or more values.
            x = data[i].strip()
            if not x: continue
            t = i
            comma = x.endswith(',') and not x.endswith(',,')
            while comma or x.endswith('\\'):
                t += 1
                if comma:
                    x += data[t].strip()
                else:
                    x = x[:-1] + data[t]
                data[t] = ''
                comma = x.endswith(',')

            prop = None
            for qtype in ('"', "'"):
                if x.startswith(qtype):
                    offset = 1
                    while True:
                        tn = x.find(qtype, offset)
                        try:
                            new_search_cond = x[tn - 1] == '\\' \
                                and x[tn - 2] != '\\'
                        except IndexError:
                            new_search_cond = False
                        if tn == -1:
                            raise SyntaxError('missing end quote \
in entry beginning on line %d' % (i + 1))
                        elif new_search_cond:
                            offset = tn + 1
                        else:
                            prop = self.convert(x[:tn + 1])
                            cval = x[tn + 1:].strip()
                            if not cval.startswith('='):
                                raise SyntaxError('no equal sign \
present in entry beginning on line %d' % (i + 1))
                            else:
                                cval = cval[1:].lstrip()
                            break

            if not prop:
                tn = x.find('=')
                prop = x[:tn].rstrip()
                cval = x[tn + 1:].strip()
            vals = []
            done = False
            while not done:
                halfdone = False
                for qtype in ('"', "'"):
                    if cval.startswith(qtype):
                        offset = 1
                        while True:
                            tn = cval.find(qtype, offset)
                            try:
                                new_search_cond = cval[tn - 1] == '\\' \
                                    and cval[tn - 2] != '\\'
                            except IndexError:
                                new_search_cond = False
                            if tn == -1:
                                raise SyntaxError('missing end quote \
in entry beginning on line %d' % (i + 1))
                            elif new_search_cond:
                                offset = tn + 1
                            else:
                                break
                        vals.append(self.convert(cval[:tn + 1]))
                        cval = cval[tn + 1:].lstrip()
                        if cval.startswith(','):
                            cval = cval[1:].lstrip()
                            halfdone = True
                            break
                        else:
                            done = True
                            break
                if done: break
                elif halfdone: continue
                # If not with quotes
                end = cval.find(',')
                if end != -1:
                    vals.append(self.convert(cval[:end].rstrip()))
                    cval = cval[end + 1:].lstrip()
                else:
                    vals.append(self.convert(cval.rstrip()))
                    break
            if len(vals) == 1:
                vals = vals[0]
            else:
                nvals = []
                for x in vals:
                    if x != '': nvals.append(x)
                vals = nvals
            self.__setitem__(prop, vals)
            

###############
## FUNCTIONS ##
###############

# Parsing (reading)
#~~~~~~~~~~~~~~~~~~
def parse(in_file=None, **kwargs):
    """
    parse([filename|file_object], (file=filename|file_object)|data=string)
    
    Parse simple config files adhering to the qvikconfig format. The
    basic syntax is 'property = value'. Executing help(qvikconfig)
    will reveal the complete documentation.
    """
    class QvikParser(_QvikParser): pass
    return QvikParser(in_file, **kwargs)

# Dumping (writing)
#~~~~~~~~~~~~~~~~~~
# Disallowed characters and patterns in strings not using
# quotes. Newlines and others must be put inside quotes, but spaces
# and tabs are okay.
_bads_general = '\n\r\a\f\b\v' + chr(0x15)
_bads_prop = _bads_general + '='    # For properties
_bads_vals = _bads_general + ',"\'' # For values
_bads_vals_regexes = [
    re.compile(r'^.*?\\$'),
    re.compile(r'^none|true|false$')
    ]

def _convert_back(dat, bads, regexes=None):
    # Convert a piece of internal Python data to a string matching the
    # rules of the qvikconfig format
    if dat is None:
        return 'none'
    if dat is True:
        return 'true'
    if dat is False:
        return 'false'
    if isinstance(dat, basestring):
        # Quote the string if it contains forbidden letters or patterns
        ok = True
        for l in dat:
            if l in bads:
                ok = False
                break
        if ok and regexes:
            for x in regexes:
                if x.match(dat):
                    ok = False
                    break
        if ok:
            return dat
        else:
            return repr(dat)
    # Else
    return str(dat) # ints and the like

def _yield_dump(in_dict):
    # Dump dict data one entry at a time
    for o_prop, o_vals in in_dict.iteritems():
        if not isinstance(o_prop, basestring):
            raise TypeError('identifiers must be strings')
        prop = _convert_back(o_prop, _bads_prop)
        if isinstance(o_vals, list):
            vals = ''
            for x in o_vals:
                vals += _convert_back(x, _bads_vals, _bads_vals_regexes) + ', '
            vals = vals[:-2]
            if len(o_vals) == 1:
                vals += ',,'
        else:
            vals = _convert_back(o_vals, _bads_vals, _bads_vals_regexes)
        yield '%s = %s\n' % (prop, vals)

def dump(in_dict, out_file=None):
    """
    dump(dict[, outfile])
    
    Translate a dict to the qvikconfig syntax. If no outfile is given,
    the output will be returned as a string.
    """
    if out_file:
        for x in _yield_dump(in_dict):
            out_file.write(x)
        return True
    else:
        newdata = ''
        for x in _yield_dump(in_dict):
            newdata += x
        return newdata



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# forbi: a TCP-based communication tool
# Copyright (C) 2010  Niels Serup

# This file is part of forbi.
#
# forbi is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# forbi is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General
# Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with forbi. If not, see
# <http://www.gnu.org/licenses/>.

##[ Maintainer  ]## Niels Serup <ns@metanohi.org>
##[ Description ]## Client communicator

import sys
import socket

from forbi.communicator import ForbiCommunicator, \
    ForbiRequestHandler, ForbiSocket
import forbi.misc as misc
import forbi.crypt as crypt
import forbi.localconn as localconn
import forbi.messagereceived as msgrecv

class ForbiClient(ForbiCommunicator):
    """A client server meant for requesting data from a remote server"""
    def __init__(self, host, port=54547, user=None, password=None,
                 **kwds):
        self.host = host
        self.port = port
        self.user = user
        self.password = password

        self.sender = None
        self.logged_in = False

        self.client_port = kwds.get('client_port') or self.port
        self.program = kwds.get('program')

        # Create the client server handler
        ForbiCommunicator.__init__(
            self, self.client_port, '', ForbiClientRequestHandler,
            kwds.get('runner'))

    def start(self):
        """Start the local server and connect to the remote server"""
        if self.host is None:
            self.host = self.runner.get_input(u'Enter host: ')
        self.set_output_prefix(u'client')
        ForbiCommunicator.start(self)
        self.client_port = self.serving_port
        self.addr_text_send = repr((self.host, self.port))
        self.state(u'creating socket to %s' % repr((self.host,
                                                   self.port)))
        self.sender = ForbiSocket(self.host, self.port, self.priv_key)
        self.identifier_string = crypt.generate_random_bytes()
        self.run()

    def send(self, msg, utf8=True):
        """Send a message to the remote server"""
        try:
            return self.sender.send(msg, utf8)
        except socket.error, e:
            self.error(u'socket error: %s' % str(e))

    def run(self):
        """Perform basic post-connection stuff"""
        self.state(u'exchanging public keys')
        pub_key_str = self.send('public_key %s' % self.pub_key_str,
                                False)
        pub_key = crypt.create_pub_key_from_string(pub_key_str)
        self.sender.pub_key = pub_key
        self.server_pub_key = pub_key
        self.sender.do_crypt = True
        self.do_crypt = True
        self.state(u'now using encryption')

        self.state(u'exchanging version numbers')
        server_version = self.send(u'version %s' %
                                   misc.version_num_text)
        if tuple(map(int, server_version.split(u'.'))) < misc.version_num:
            self.error(u'versions incompatible')
            sys.exit()
        self.state(u'client using v%s' % misc.version_num_text)
        self.state(u'server using v%s' % server_version)
        self.state(u'versions compatible')

        self.state(u'sending identifier string to server')
        if self.send('id %s' % self.identifier_string, False) != u'ok':
            self.error(u'not ok')
            return

        self.state(u'sending port number to server')
        if self.send(u'port %s' % self.client_port) != u'ok':
            self.error(u'not ok')
            return

        if self.user is None or self.password is None:
            self.state(u'creating user...')
            if self.user is None:
                while True:
                    user = self.runner.get_input(u'Enter user name: ')
                    if user == '':
                        self.error(u'username cannot be empty')
                    elif '\t' in user:
                        self.error(u'username cannot contain tabs')
                    else:
                        break
                self.user = user
            if self.password is None:
                while True:
                    password = self.runner.get_input(u'Enter password: ', False)
                    if password == '':
                        self.error(u'password cannot be empty')
                        continue
                    password_check = self.runner.get_input(u'Reenter password: ', False)
                    if password != password_check:
                        self.error(u'the two passwords do not match')
                    else:
                        break
                self.password = password
        try:
            self.login()
        except misc.ForbiError:
            return
        
        self.state(u'asking for list of online users')
        users = self.send(u'users')
        if users.startswith(u'ok '):
            users = u', '.join(repr(x)[1:] for x in users.split(u' ', 1)[1].split(u'\t'))
        else:
            users = u'none'
        self.state(u'users online: %s' % users)

        self.local_connection_server = localconn.LocalServer(self)
        self.local_connection_server.start()
        misc.QuickThread(self.local_connection_server.run)
        ForbiCommunicator.run(self)

    def receive_local_command(self, all_args):
        """Receive command from computer via D-Bus and act on it"""
        command = all_args[0]
        args = all_args[1:]

        if command == u'message':
            # Receive and show message
            if not self.logged_in:
                return u'error: not logged in'
            recp = repr(args[0])[1:]
            self.state(u'sending message to %s' % recp)
            msg = u'%s %s\t%s' % tuple(all_args)
            ans = self.send(msg)
            if ans == u'ok':
                self.state(u'message to %s succesfully sent' % recp)
                return u'ok'
            elif ans.startswith(u'error'):
                if u' ' in ans:
                    extra = u': ' + ans.split(u' ', 1)[1]
                else:
                    extra = ''
                self.error(u'message to %s not sent%s' % (recp, extra))
                return u'error%s' % extra
            else:
                self.error(u'not ok')
                return u'not ok'
                raise misc.ForbiError

        elif command == u'users':
            # Receive list of users
            ans = self.send(u'users')
            if ans.startswith(u'ok '):
                ans = ans.split(u' ', 1)[1]
                if ans:
                    return u', '.join([repr(u)[1:] for u in ans.split('\t')])
                else:
                    return u'none'
            else:
                return u'error: not logged in'

        elif command == u'logout':
            # Logout
            try:
                self.logout()
                return u'ok'
            except misc.ForbiError:
                return u'error: %s' % self.runner.prev_error

        elif command == u'exit':
            # Exit completely
            self.exit()
            return u'ok'

        elif command == u'login':
            # Login (and register if necessary)
            if self.logged_in:
                return u'error: currently logged in as %s' % repr(self.user)[1:]
            else:
                self.user, self.password = args
                try:
                    self.login()
                    return u'ok'
                except misc.ForbiError:
                    return u'error: %s' % self.runner.prev_error

        elif command == u'register':
            # Register (but don't login)
            if self.logged_in:
                return u'error: currently logged in as %s' % repr(self.user)[1:]
            else:
                self.user, self.password = args
                try:
                    self.register()
                    return u'ok'
                except misc.ForbiError:
                    return u'error: %s' % self.runner.prev_error

        elif command == u'unregister':
            # Unregister
            try:
                self.unregister()
                return u'ok'
            except misc.ForbiError:
                return u'error: %s' % self.runner.prev_error

    def login(self):
        """Attempt to login to remote server"""
        self.state(u'attempting to log in...')
        ans = self.send(u'login %s\t%s' % (self.user, self.password))
        if ans.startswith(u'error'):
            if u' ' in ans:
                errnum = int(ans.split(u' ', 1)[1])
            else:
                errnum = 0
            if errnum == 1:
                self.error(u'user is already logged in on server')
                raise misc.ForbiError
            elif errnum == 2:
                self.error(u'incorrect password')
                raise misc.ForbiError
            elif errnum == 3:
                self.error(u'user does not exist on server')
                self.register()
                self.login()
            else:
                self.error(u'not ok')
                raise misc.ForbiError
        else:
            self.state(u'succesfully logged in')
            self.logged_in = True

    def register(self):
        """Attempt to register user on the server"""
        self.state(u'attempting to register...')
        ans = self.send(u'register %s\t%s' % (self.user,
                                              self.password))
        if ans.startswith(u'error'):
            if u' ' in ans:
                errnum = int(ans.split(u' ', 1)[1])
            else:
                errnum = 0
            if errnum == 1:
                self.error(u'user is already registered on server')
                raise misc.ForbiError
            else:
                self.error(u'not ok')
                raise misc.ForbiError
        else:
            self.state(u'succesfully registered')

    def logout(self):
        """Logout from remote server"""
        self.state(u'logging out...')
        if not self.logged_in:
            self.error(u'not logged in')
            raise misc.ForbiError
        ans = self.send(u'logout')
        if ans == u'ok':
            self.state(u'logged out')            
            self.logged_in = False
        else:
            self.error(u'not ok')
            raise misc.ForbiError

    def unregister(self):
        """Unregister on remote server"""
        self.state(u'unregistering...')
        if not self.logged_in:
            self.error(u'not logged in')
            raise misc.ForbiError
        ans = self.send(u'unregister')
        if ans == u'ok':
            self.state(u'unregistered')
            self.logged_in = False
            self.user = None
            self.password = None
        else:
            self.error(u'not ok')
            raise misc.ForbiError

    def exit(self):
        """Send an exit message to the server and quit"""
        self.state(u'exiting...')
        self.send(u'exit')
        self.end()

    def end(self):
        """Stop the sender socket and end everything"""
        self.end = misc.nothing
        try:
            self.exit()
        except misc.ForbiError:
            pass
        if self.sender is not None:
            self.sender.close()
        ForbiCommunicator.end(self)

class ForbiClientRequestHandler(ForbiRequestHandler):
    """A request handler with support for ponging and receiving messages"""
    def start(self):
        """Finalize the connection"""
        self.addr_text_recv = repr(self.client_address)
        self.is_server = None
        msgrecv.init(self.parent.program)

    def end(self):
        if self.is_server:
            self.parent.end()

    def receive(self, msg):
        """
        Receive and check identification strings and, if the server
        succesfully identifies itself as being the server, continue
        with receiving other commands.
        """
        if msg.startswith('id'):
            if ' ' in msg and msg.split(' ', 1)[1] == \
                    self.parent.identifier_string:
                self.state(u'received matching identifier string from server')
                self.is_server = True
                self.receive = self.server_receive
                self.answer(u'ok')
            else:
                if msg == 'id':
                    self.error(u'received empty identifier string from server')
                else:
                    self.error(u'received non-matching identifier string from server')
                self.is_server = False
                self.receive = self.foreign_client_receive

    def server_receive(self, msg):
        """Receive a message from a remote server and act on it"""
        if msg == u'ping':
            # Ping.. pong..
            self.state(u'received ping from server. Ponging.')
            self.answer(u'pong')
        elif msg.startswith(u'message '):
            # Show message using either internal GUI or external program
            user_from, mesg = msg.split(u' ', 1)[1].split(u'\t')
            self.state(u'message received from server. Showing.')
            msgrecv.new_message(user_from, self.parent.user, mesg)
            self.answer(u'ok')

    def foreign_client_receive(self, msg):
        # Foreign clients aren't meant to contact other clients
        # directly, so this is left out intentionally, not because I
        # was too lazy to implement it. Nevertheless, it could work.
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# forbi: a TCP-based communication tool
# Copyright (C) 2010  Niels Serup

# This file is part of forbi.
#
# forbi is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# forbi is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General
# Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with forbi. If not, see
# <http://www.gnu.org/licenses/>.

##[ Maintainer  ]## Niels Serup <ns@metanohi.org>
##[ Description ]## Miscellaneous data

# Basic program data
program_name = u'forbi'
program_sender_name = u'forbi-send'
program_description = u'A TCP-based communication tool'
program_sender_description = u'Send messages to foreign forbi clients via a local running client'
version_num = (0, 1, 0)
version_num_text = u'.'.join([unicode(x) for x in version_num])
program_author = u'Niels Serup'
author_email = u'ns@metanohi.org'
version_text = u'''\
%s %s
Copyright (C) 2010  %s
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>
This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law.\
''' % (program_name, version_num_text, program_author)

# Basic program functions and classes
import sys
import locale
import datetime
import threading
import subprocess
try:
    from termcolor import colored
except ImportError:
    from forbi.external.termcolor import colored

# Get the system encoding
encoding = locale.getpreferredencoding()
    
def output_text(msg, out, color=None, prefix=''):
    """Output text with or without color, with or without a prefix"""
    if prefix:
        prefix += ': '
    date = datetime.datetime.today().strftime('%T')
    msg = '%s[%s]: %s%s' % (program_name, date, prefix, msg)
    msg = msg.encode(encoding)
    if color:
        msg = colored(msg, color)
    out.write(msg + '\n')

def state(msg, color=True, prefix='', out=sys.stdout):
    """State something"""
    output_text(msg, out, color and 'blue', prefix)
    
def error(msg, color=True, prefix='', out=sys.stderr):
    """Output an error message"""
    output_text(msg, out, color and 'red', prefix)

nothing = lambda *args, **kwargs: None

class TextOutputterBaseFallback(object):
    def state(self, msg, prefix):
        state(msg, False, prefix)

    def error(self, msg, prefix):
        error(msg, False, prefix)

_textoutputter_base_fallback = TextOutputterBaseFallback()
        
class TextOutputter(object):
    """A text outputter"""
    def __init__(self, base=None, prefix=''):
        self.set_textoutputter_base(base)
        self.set_output_prefix(prefix)

    def set_textoutputter_base(self, obj):
        """Set the base object whose functions should be called"""
        self.textoutputter_base = obj or _textoutputter_base_fallback

    def set_output_prefix(self, prefix):
        """Set the output prefix"""
        self.output_prefix = prefix
        
    def state(self, msg):
        """State something via the base object"""
        self.textoutputter_base.state(msg, self.output_prefix)

    def error(self, msg):
        """Output an error message via the base object"""
        self.textoutputter_base.error(msg, self.output_prefix)

class BasicQuickThread(threading.Thread):
    """A quick thread creator and runner"""
    def __init__(self, func, *args, **kwds):
        threading.Thread.__init__(self)
        self.func = func
        self.args = args
        self.kwds = kwds
        self.do_terminate = False
        self.setDaemon(True) # self.daemon = True

    def run(self):
        self.func(*self.args, **self.kwds)

    def terminate(self):
        self.do_terminate = True

class QuickThread(BasicQuickThread):
    """A quicker thread creator and runner"""
    def __init__(self, *args, **kwds):
        BasicQuickThread.__init__(self, *args, **kwds)
        self.start()

class Container(object):
    """A container top class"""

class ForbiError(Exception):
    pass

def exec_program(*args):
    """Run a program in the background"""
    return subprocess.Popen(map(str, args)).pid



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# forbi: a TCP-based communication tool
# Copyright (C) 2010  Niels Serup

# This file is part of forbi.
#
# forbi is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# forbi is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General
# Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with forbi. If not, see
# <http://www.gnu.org/licenses/>.

# This file is based on an example by Red Hat Inc. and Collabora Ltd.
#
# Copyright (C) 2004-2006 Red Hat Inc. <http://www.redhat.com/>
# Copyright (C) 2005-2007 Collabora Ltd. <http://www.collabora.co.uk/>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the 'Software'), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

##[ Maintainer  ]## Niels Serup <ns@metanohi.org>
##[ Description ]## D-Bus specifics

import dbus
import dbus.service
import dbus.mainloop.glib
import gobject
import traceback

# Make sure threads work nicely together with gobject
gobject.threads_init()

class LocalServerObject(dbus.service.Object):
    def __init__(self, parent, *args):
        self.parent = parent
        dbus.service.Object.__init__(self, *args)

    # Receive an array of strings (as) and return a string (s)
    @dbus.service.method('org.forbi', in_signature='as', out_signature='s')
    def request(self, args):
        args = map(unicode, args)
        return self.parent.receive(args)

class LocalServer(object):
    """A D-Bus server listening for commands sent by D-Bus clients"""
    def __init__(self, parent):
        self.parent = parent
    
    def start(self):
        """Prepare the D-Bus loop"""
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

        self.session_bus = dbus.SessionBus()
        self.name = dbus.service.BusName('org.forbi', self.session_bus)
        self.object = LocalServerObject(self, self.session_bus, '/Server')

        self.mainloop = gobject.MainLoop()

    def receive(self, args):
        """
        Receive commands and forward them to this server object's
        parent object
        """
        try:
            return self.parent.receive_local_command(args)
        except Exception:
            traceback.print_exc()

    def run(self):
        """Start the D-Bus loop"""
        self.mainloop.run()

class LocalClient(object):
    """A D-Bus client sending commands to a D-Bus server"""
    def __init__(self):
        self.bus = dbus.SessionBus()
        # Connect to the server
        self.remote_object = self.bus.get_object('org.forbi', '/Server')
        self.interface = dbus.Interface(self.remote_object, 'org.forbi')

    def send(self, args):
        """Send a command -- an array of strings -- to the server"""
        reply = unicode(self.interface.request(args))
        return reply

if __name__ == '__main__':
    # A simple server
    class SimpleParent(object):
        def receive_local_command(self, args):
            print 'received', repr(args)
    s = LocalServer(SimpleParent())
    s.start()
    s.run()



#!/usr/bin/env python
from distutils.core import setup
import sys
import os
import fnmatch

basedir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, basedir)
import forbi.misc as misc

readme = open('README.txt').read()
conf = dict(
    name=misc.program_name,
    version=misc.version_num_text,
    description=misc.program_description,
    author=misc.program_author,
    author_email=misc.author_email,
    packages=['forbi', 'forbi.external'],
    scripts=['scripts/forbi', 'scripts/forbi-send'],
    requires=['M2Crypto', 'dbus', 'gobject', 'tk'],
    url='http://metanohi.org/projects/forbi/',
    license='AGPLv3+',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'License :: DFSG approved',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Communications',
        'Topic :: Internet',
        'Topic :: Utilities'
        ]
)

try:
    # setup.py register wants unicode data..
    conf['long_description'] = readme.decode('utf-8')
    setup(**conf)
except Exception:
    # ..but setup.py sdist upload wants byte data
    conf['long_description'] = readme
    setup(**conf)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# forbi: a TCP-based communication tool
# Copyright (C) 2010  Niels Serup

# This file is part of forbi.
#
# forbi is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# forbi is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General
# Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with forbi. If not, see
# <http://www.gnu.org/licenses/>.

##[ Maintainer  ]## Niels Serup <ns@metanohi.org>
##[ Description ]## Cryptography-related functions

import tempfile
import M2Crypto
import base64

def generate_keys():
    """Generate a 1024-bit keypair"""
    priv = M2Crypto.RSA.gen_key(1024, 65537)
    tf = tempfile.NamedTemporaryFile()
    priv.save_pub_key(tf.name)
    pub = M2Crypto.RSA.load_pub_key(tf.name)
    tf.seek(0)
    pub_str = tf.read()
    return priv, pub, pub_str

def create_pub_key_from_string(string):
    """Transform a string in the PEM format into a public key object"""
    tf = tempfile.NamedTemporaryFile()
    tf.write(string)
    tf.seek(0)
    pub = M2Crypto.RSA.load_pub_key(tf.name)
    return pub

def encrypt(msg, pub):
    """Encrypt plaintext with a public key"""
    padding = M2Crypto.RSA.pkcs1_oaep_padding
    ctxt=pub.public_encrypt(msg, padding)
    return ctxt

def decrypt(msg, priv):
    """Decrypt ciphertext with a private key"""
    padding = M2Crypto.RSA.pkcs1_oaep_padding
    ctxt=priv.private_decrypt(msg, padding)
    return ctxt

def generate_random_bytes():
    """Generate 128 bytes of pseudo-randomness"""
    return base64.b64encode(M2Crypto.Rand.rand_bytes(128))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# forbi: a TCP-based communication tool
# Copyright (C) 2010  Niels Serup

# This file is part of forbi.
#
# forbi is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# forbi is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General
# Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with forbi. If not, see
# <http://www.gnu.org/licenses/>.

##[ Maintainer  ]## Niels Serup <ns@metanohi.org>
##[ Description ]## Runner of client or server

import sys
import os
import cStringIO as StringIO
import getpass
from forbi.settingsparser import SettingsParser
import forbi.misc as misc
from forbi.client import ForbiClient
from forbi.server import ForbiServer

_config_file_translations = {
    'save file': 'save_file',
    'execute on new message': 'execute_on_new_message',
    'verbose': 'term_verbose',
    'colored text': 'term_colored_text',
    'server': 'is_server',
    'host': 'server_host',
    'port': 'server_port',
    'user': 'server_user',
    'password': 'server_password',
    'client port': 'client_port'
}

class ForbiRunner(SettingsParser):
    """
    A root class containing base functions meant for use by class
    objects
    """
    def __init__(self, **options):
        SettingsParser.__init__(self, _config_file_translations,
                                **options)
        # Get values, or set them to default ones
        self.set_if_nil('use_config_file', True)
        if self.use_config_file:
            self.set_if_nil('config_file_path',
                            os.path.expanduser('~/.forbi.conf'))
            self.load_from_file()
        self.set_if_nil('save_file', '~/.forbi.db')
        self.save_file = os.path.expanduser(self.save_file)
        self.set_if_nil('execute_on_new_message', None)

        self.set_if_nil('term_verbose', True)
        self.set_if_nil('term_colored_text', True)
        self.set_if_nil('is_server', False)
        self.set_if_nil('server_host', None)
        self.set_if_nil('server_port', 54547)
        self.set_if_nil('server_user', None)
        self.set_if_nil('server_password', None)
        self.set_if_nil('client_port', -1) # -1 = pick a random port

        self.communicator = None
        self.text_output_paused = False
        self.prev_error = None
        self.prev_state = None

    def state(self, msg, prefix=''):
        """State something"""
        if self.term_verbose:
            if self.text_output_paused:
                out = self.state_text_cache
            else:
                out = sys.stdout
            misc.state(msg, self.term_colored_text, prefix, out)
        self.prev_state = msg

    def error(self, msg, prefix=''):
        """Output an error message"""
        if self.term_verbose:
            if self.text_output_paused:
                out = self.error_text_cache
            else:
                out = sys.stderr
            misc.error(msg, self.term_colored_text, prefix)
        self.prev_error = msg

    def pause_text_output(self, pause=True):
        """Pause text output"""
        self.text_output_paused = pause
        self.state_text_cache = StringIO.StringIO()
        self.error_text_cache = StringIO.StringIO()

    def unpause_text_output(self, unpause=True):
        """Unpause text output"""
        self.flush_caches()
        self.pause_text_output(not unpause)

    def flush_caches(self):
        """Output all cached text"""
        sys.stdout.write(self.state_text_cache.getvalue())
        sys.stderr.write(self.error_text_cache.getvalue())

    def get_input(self, prompt, echo=True):
        """Get text input from the terminal"""
        self.pause_text_output()
        if echo:
            inp = raw_input(prompt)
        else:
            inp = getpass.getpass(prompt)
        self.unpause_text_output()
        return inp.decode(misc.encoding)

    def start(self):
        """
        Continue running either as server or client, depending on
        the options sent to the runner object
        """
        if not self.is_server:
            self.communicator = ForbiClient(
                self.server_host, self.server_port,
                self.server_user, self.server_password,
                runner=self, client_port=self.client_port,
                program=self.execute_on_new_message)
        else:
            self.communicator = ForbiServer(
                self.server_port, runner=self,
                save_file=self.save_file,
                default_client_port=self.client_port)

        self.communicator.start()

    def end(self):
        """End everything"""
        if self.communicator is not None:
            self.communicator.end()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# forbi: a TCP-based communication tool
# Copyright (C) 2010  Niels Serup

# This file is part of forbi.
#
# forbi is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# forbi is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General
# Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with forbi. If not, see
# <http://www.gnu.org/licenses/>.

##[ Maintainer  ]## Niels Serup <ns@metanohi.org>
##[ Description ]## Generic classes for sending and receiving data

import socket
import SocketServer
import threading
import time
import struct
import random
import traceback

import forbi.crypt as crypt
import forbi.misc as misc

class SenderReceiverEncoderDecoder(object):
    """Encodes when sending, decodes when receiving"""
    def encode(self, msg, pub_key=None, do_encrypt=False, utf8=True):
        """Encode and encrypt a message"""
        if utf8:
            msg = msg.encode('utf-8')
            msg += '1'
        else:
            msg += '0'
        if do_encrypt:
            # Creates 128-byte blocks, each with 86 or less characters
            # of the full message.
            omsg = msg
            msg = ''
            while omsg:
                # PKCS#1 requires 11 bytes of the RSA modulus, but
                # OAEP requires 41. 1024/8 - 41 = 87 bytes left.
                msg += crypt.encrypt(omsg[:86], pub_key)
                omsg = omsg[86:]
        len_str = struct.pack('<I', len(msg) + 4)
        msg = len_str + msg
        return msg

    def decode(self, recv, priv_key=None, do_decrypt=False):
        """Decode and decrypt a message"""
        msg = recv(4096)
        try:
            msg_len = struct.unpack('<I', msg[:4])[0]
        except struct.error:
            raise misc.ForbiError('lost connection')
        msg = msg[4:]
        idx = 4096
        while idx < msg_len:
            msg += recv(4096)
            idx += 4096
        if do_decrypt:
            omsg = msg
            msg = ''
            while omsg:
                # 1024 / 8 = 128
                msg += crypt.decrypt(omsg[:128], priv_key)
                omsg = omsg[128:]
        utf8 = bool(int(msg[-1:]))
        msg = msg[:-1]
        if utf8:
            msg = msg.decode('utf-8')
        return msg

class ForbiSocket(SenderReceiverEncoderDecoder):
    """
    A high level socket class capable of receiving and sending large
    amounts of data in one go with or without encryption
    """
    def __init__(self, host, port=54547, priv_key=None):
        self.host = host
        self.port = port
        self.addr = (self.host, self.port)
        self.pub_key = None
        self.priv_key = priv_key
        self.do_crypt = False

        # Create a socket
        self.socket = socket.socket(socket.AF_INET,
                                    socket.SOCK_STREAM)
        # Connect to it
        self.socket.connect(self.addr)

    def send(self, msg, utf8=True):
        """Send a message"""
        msg = self.encode(msg, self.pub_key, self.do_crypt, utf8)
        self.socket.sendall(msg)
        return self.recv()

    def recv(self):
        """Receive an answer"""
        return self.decode(self.socket.recv, self.priv_key, self.do_crypt)

    def close(self):
        self.socket.close()

    def end(self):
        self.close()

class BasicThreadedTCPServer(SocketServer.ThreadingMixIn,
                             SocketServer.TCPServer):
    pass

class ThreadedTCPServer(BasicThreadedTCPServer):
    def __init__(self, addr, handler, parent):
        self.parent = parent
        BasicThreadedTCPServer.__init__(self, addr, handler)

class ForbiRequestHandler(SocketServer.BaseRequestHandler,
                          SenderReceiverEncoderDecoder):
    """A base request handler"""
    def setup(self):
        # Setup variables for when new requests come in
        self.parent = self.server.parent
        self.priv_key = None
        self.pub_key = None
        self.do_crypt = False
        self.state = self.parent.state
        self.error = self.parent.error
        self.state(u'connected to %s' % repr(self.client_address))
        self.start()

    def handle(self):
        # Continously handle incoming messages
        self.state(u'started handling requests from %s' % repr(self.client_address))
        while True:
            try:
                msg = self.recv()
                if not msg:
                    raise Exception(u'no more messages')
                if self.receive(msg):
                    break
            except socket.error:
                break
            except Exception, e:
                self.error(u'handling error: %s' % repr(e))
#                traceback.print_exc()
                break
        self.state(u'stopped handling requests from %s' % repr(self.client_address))
        self.end()

    def receive(self, msg):
        """
        Is run when a new message is received (should be overridden by
        subclasses)
        """

    def answer(self, msg, utf8=True):
        self.request.sendall(self.encode(
                msg, self.pub_key or self.parent.server_pub_key,
                self.do_crypt or self.parent.do_crypt, utf8))

    def recv(self):
        return self.decode(self.request.recv, self.priv_key or
                           self.parent.priv_key, self.do_crypt
                           or self.parent.do_crypt)

    def start(self):
        """Is run after setup (should be overridden by subclasses)"""

    def end(self):
        """
        Is run when there is nothing more to handle (should be
        overridden by subclasses)
        """

class ForbiCommunicator(misc.TextOutputter):
    """A commmunicator server receiving and answering messages"""
    def __init__(self, serving_port=54547, server_host='',
                 handler=None, runner=None):
        self.serving_port = serving_port
        self.server_host = server_host
        self.server = None
        self.handler = handler or ForbiRequestHandler
        self.pub_key = None
        self.priv_key = None
        self.runner = runner
        self.quit_program = False
        misc.TextOutputter.__init__(self, self.runner)
        
    def start(self):
        """Start the communicator"""
        self.state(u'generating keys...')
        self.do_crypt = False
        self.generate_keys()
        self.server_pub_key = None

        self.state(u'creating server...')

        # Create a threaded TCP server
        if self.serving_port != -1:
            self.server = ThreadedTCPServer((self.server_host, self.serving_port),
                                            self.handler, self)
        else:
            while True:
                serving_port = random.randint(49152, 65536)
                try:
                    self.server = ThreadedTCPServer(
                        (self.server_host, serving_port),
                        self.handler, self)
                    break
                except socket.error, e:
                    if e[0] == 98: # "Address already in use"
                        continue
                    else:
                        traceback.print_exc()
                        break
            self.serving_port = serving_port
        
        self.server_ip, self.serving_port = self.server.server_address

        # Run the server in a thread. This thread will start a new
        # thread for every client connection request
        self.server_thread = threading.Thread(target=self.server.serve_forever)

        # Exit the server thread when the main thread terminates
        self.server_thread.setDaemon(True)
        self.server_thread.start()

        self.state('serving on port %d' % self.serving_port)

    def generate_keys(self):
        """Generate keypair"""
        self.priv_key, self.pub_key, self.pub_key_str = \
            crypt.generate_keys()

    def run(self, extra_func=None):
        """Run communicator server without end"""
        if extra_func is None:
            extra_func = lambda: time.sleep(.5)
        self.state(u'running indefinitely...')
        while not self.quit_program:
            extra_func()

    def end(self):
        """End everything"""
        self.state(u'shutting down...')
        if self.server is not None:
            self.server.shutdown()
            self.quit_program = True
        self.end = misc.nothing
